# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cast_channel.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cast_channel.proto',
  package='extensions.api.cast_channel',
  syntax='proto2',
  serialized_options=b'H\003',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x12\x63\x61st_channel.proto\x12\x1b\x65xtensions.api.cast_channel\"\xe3\x02\n\x0b\x43\x61stMessage\x12R\n\x10protocol_version\x18\x01 \x02(\x0e\x32\x38.extensions.api.cast_channel.CastMessage.ProtocolVersion\x12\x11\n\tsource_id\x18\x02 \x02(\t\x12\x16\n\x0e\x64\x65stination_id\x18\x03 \x02(\t\x12\x11\n\tnamespace\x18\x04 \x02(\t\x12J\n\x0cpayload_type\x18\x05 \x02(\x0e\x32\x34.extensions.api.cast_channel.CastMessage.PayloadType\x12\x14\n\x0cpayload_utf8\x18\x06 \x01(\t\x12\x16\n\x0epayload_binary\x18\x07 \x01(\x0c\"!\n\x0fProtocolVersion\x12\x0e\n\nCASTV2_1_0\x10\x00\"%\n\x0bPayloadType\x12\n\n\x06STRING\x10\x00\x12\n\n\x06\x42INARY\x10\x01\"\x0f\n\rAuthChallenge\"B\n\x0c\x41uthResponse\x12\x11\n\tsignature\x18\x01 \x02(\x0c\x12\x1f\n\x17\x63lient_auth_certificate\x18\x02 \x02(\x0c\"~\n\tAuthError\x12\x44\n\nerror_type\x18\x01 \x02(\x0e\x32\x30.extensions.api.cast_channel.AuthError.ErrorType\"+\n\tErrorType\x12\x12\n\x0eINTERNAL_ERROR\x10\x00\x12\n\n\x06NO_TLS\x10\x01\"\xc6\x01\n\x11\x44\x65viceAuthMessage\x12=\n\tchallenge\x18\x01 \x01(\x0b\x32*.extensions.api.cast_channel.AuthChallenge\x12;\n\x08response\x18\x02 \x01(\x0b\x32).extensions.api.cast_channel.AuthResponse\x12\x35\n\x05\x65rror\x18\x03 \x01(\x0b\x32&.extensions.api.cast_channel.AuthErrorB\x02H\x03'
)



_CASTMESSAGE_PROTOCOLVERSION = _descriptor.EnumDescriptor(
  name='ProtocolVersion',
  full_name='extensions.api.cast_channel.CastMessage.ProtocolVersion',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CASTV2_1_0', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=335,
  serialized_end=368,
)
_sym_db.RegisterEnumDescriptor(_CASTMESSAGE_PROTOCOLVERSION)

_CASTMESSAGE_PAYLOADTYPE = _descriptor.EnumDescriptor(
  name='PayloadType',
  full_name='extensions.api.cast_channel.CastMessage.PayloadType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STRING', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BINARY', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=370,
  serialized_end=407,
)
_sym_db.RegisterEnumDescriptor(_CASTMESSAGE_PAYLOADTYPE)

_AUTHERROR_ERRORTYPE = _descriptor.EnumDescriptor(
  name='ErrorType',
  full_name='extensions.api.cast_channel.AuthError.ErrorType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INTERNAL_ERROR', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NO_TLS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=577,
  serialized_end=620,
)
_sym_db.RegisterEnumDescriptor(_AUTHERROR_ERRORTYPE)


_CASTMESSAGE = _descriptor.Descriptor(
  name='CastMessage',
  full_name='extensions.api.cast_channel.CastMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='protocol_version', full_name='extensions.api.cast_channel.CastMessage.protocol_version', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_id', full_name='extensions.api.cast_channel.CastMessage.source_id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='destination_id', full_name='extensions.api.cast_channel.CastMessage.destination_id', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='extensions.api.cast_channel.CastMessage.namespace', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload_type', full_name='extensions.api.cast_channel.CastMessage.payload_type', index=4,
      number=5, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload_utf8', full_name='extensions.api.cast_channel.CastMessage.payload_utf8', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload_binary', full_name='extensions.api.cast_channel.CastMessage.payload_binary', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CASTMESSAGE_PROTOCOLVERSION,
    _CASTMESSAGE_PAYLOADTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=52,
  serialized_end=407,
)


_AUTHCHALLENGE = _descriptor.Descriptor(
  name='AuthChallenge',
  full_name='extensions.api.cast_channel.AuthChallenge',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=409,
  serialized_end=424,
)


_AUTHRESPONSE = _descriptor.Descriptor(
  name='AuthResponse',
  full_name='extensions.api.cast_channel.AuthResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signature', full_name='extensions.api.cast_channel.AuthResponse.signature', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_auth_certificate', full_name='extensions.api.cast_channel.AuthResponse.client_auth_certificate', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=426,
  serialized_end=492,
)


_AUTHERROR = _descriptor.Descriptor(
  name='AuthError',
  full_name='extensions.api.cast_channel.AuthError',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='error_type', full_name='extensions.api.cast_channel.AuthError.error_type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _AUTHERROR_ERRORTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=494,
  serialized_end=620,
)


_DEVICEAUTHMESSAGE = _descriptor.Descriptor(
  name='DeviceAuthMessage',
  full_name='extensions.api.cast_channel.DeviceAuthMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='challenge', full_name='extensions.api.cast_channel.DeviceAuthMessage.challenge', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='response', full_name='extensions.api.cast_channel.DeviceAuthMessage.response', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error', full_name='extensions.api.cast_channel.DeviceAuthMessage.error', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=623,
  serialized_end=821,
)

_CASTMESSAGE.fields_by_name['protocol_version'].enum_type = _CASTMESSAGE_PROTOCOLVERSION
_CASTMESSAGE.fields_by_name['payload_type'].enum_type = _CASTMESSAGE_PAYLOADTYPE
_CASTMESSAGE_PROTOCOLVERSION.containing_type = _CASTMESSAGE
_CASTMESSAGE_PAYLOADTYPE.containing_type = _CASTMESSAGE
_AUTHERROR.fields_by_name['error_type'].enum_type = _AUTHERROR_ERRORTYPE
_AUTHERROR_ERRORTYPE.containing_type = _AUTHERROR
_DEVICEAUTHMESSAGE.fields_by_name['challenge'].message_type = _AUTHCHALLENGE
_DEVICEAUTHMESSAGE.fields_by_name['response'].message_type = _AUTHRESPONSE
_DEVICEAUTHMESSAGE.fields_by_name['error'].message_type = _AUTHERROR
DESCRIPTOR.message_types_by_name['CastMessage'] = _CASTMESSAGE
DESCRIPTOR.message_types_by_name['AuthChallenge'] = _AUTHCHALLENGE
DESCRIPTOR.message_types_by_name['AuthResponse'] = _AUTHRESPONSE
DESCRIPTOR.message_types_by_name['AuthError'] = _AUTHERROR
DESCRIPTOR.message_types_by_name['DeviceAuthMessage'] = _DEVICEAUTHMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CastMessage = _reflection.GeneratedProtocolMessageType('CastMessage', (_message.Message,), {
  'DESCRIPTOR' : _CASTMESSAGE,
  '__module__' : 'cast_channel_pb2'
  # @@protoc_insertion_point(class_scope:extensions.api.cast_channel.CastMessage)
  })
_sym_db.RegisterMessage(CastMessage)

AuthChallenge = _reflection.GeneratedProtocolMessageType('AuthChallenge', (_message.Message,), {
  'DESCRIPTOR' : _AUTHCHALLENGE,
  '__module__' : 'cast_channel_pb2'
  # @@protoc_insertion_point(class_scope:extensions.api.cast_channel.AuthChallenge)
  })
_sym_db.RegisterMessage(AuthChallenge)

AuthResponse = _reflection.GeneratedProtocolMessageType('AuthResponse', (_message.Message,), {
  'DESCRIPTOR' : _AUTHRESPONSE,
  '__module__' : 'cast_channel_pb2'
  # @@protoc_insertion_point(class_scope:extensions.api.cast_channel.AuthResponse)
  })
_sym_db.RegisterMessage(AuthResponse)

AuthError = _reflection.GeneratedProtocolMessageType('AuthError', (_message.Message,), {
  'DESCRIPTOR' : _AUTHERROR,
  '__module__' : 'cast_channel_pb2'
  # @@protoc_insertion_point(class_scope:extensions.api.cast_channel.AuthError)
  })
_sym_db.RegisterMessage(AuthError)

DeviceAuthMessage = _reflection.GeneratedProtocolMessageType('DeviceAuthMessage', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEAUTHMESSAGE,
  '__module__' : 'cast_channel_pb2'
  # @@protoc_insertion_point(class_scope:extensions.api.cast_channel.DeviceAuthMessage)
  })
_sym_db.RegisterMessage(DeviceAuthMessage)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python

import setuptools

setuptools.setup(
    name='dmcast',
    packages=['dmcast'],
    version='0.0.4',
    license='EPL',
    description='simple chromecast utility',
    url='https://github.com/xlfe/dmcast',
    keywords=[
        'python', 'chromecast', 'internet-of-things',
        'chromecast-audio', 'cast'],
    install_requires=[
        'protobuf>=2.6.0'
    ],
    classifiers=[
     'Development Status :: 3 - Alpha',
     'Intended Audience :: Developers',
     'License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)',
     'Programming Language :: Python :: 3.9',
    ]
)




#!/usr/bin/env python

# Based on work by Fred Clift from 2014
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE":
# <fred@clift.org> wrote this file. As long as you retain this notice you
# can do whatever you want with this stuff. If we meet some day, and you think
# this stuff is worth it, you can buy me a beer in return - Fred Clift
# (except that I dont drink....)
# ----------------------------------------------------------------------------
#

from . import cast_channel_pb2
import collections
import ssl
import asyncio
from struct import pack, unpack
import json

DEFAULT_APP = "CC1AD845"

namespace = {'con':        'urn:x-cast:com.google.cast.tp.connection',
             'receiver':   'urn:x-cast:com.google.cast.receiver',
             'cast':       'urn:x-cast:com.google.cast.media',
             'heartbeat':  'urn:x-cast:com.google.cast.tp.heartbeat',
             'message':    'urn:x-cast:com.google.cast.player.message',
             'media':      "urn:x-cast:com.google.cast.media"}


def make_msg(ns=None):
    msg = cast_channel_pb2.CastMessage()
    msg.protocol_version = msg.CASTV2_1_0
    msg.source_id = "sender-0"
    msg.destination_id = "receiver-0"
    msg.payload_type = cast_channel_pb2.CastMessage.STRING
    if ns:
        msg.namespace = namespace[ns]
    return msg


class Chromecast():

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.running = True
        self.id_counts = collections.Counter()
        self.dmr_session = None

        self.dispatch = {
                'RECEIVER_STATUS': self.handle_receiver_status,
                'MEDIA_STATUS': self.handle_media_status,
                'PING': self.pong
                }

    async def handle_media_status(self, data):

        if data['status']:
            self.msid = data['status'][0]['mediaSessionId']

    async def handle_receiver_status(self, msg):

        state = {'namespace': 'receiver',
                 'volume_level': msg['status']['volume']['level'],
                 'volume_muted': msg['status']['volume']['muted']}

        if 'applications' in msg['status']:

            apps = msg['status']['applications']
            state['current_app'] = apps[0]['displayName']

            try:
                # find details for DMR if it is already running
                da = min(filter(lambda a: a['appId'] == DEFAULT_APP, apps))
                self.dmr_session = da['sessionId']

                # open a connection
                await self.connect(dst=self.dmr_session)

                # check if anything is playing
                await self.media_status()

            except ValueError:
                self.dmr_session = None
        else:
            self.dmr_session = None

        await self.notify_state(state)

    async def notify_state(self, state):
        pass

    async def recv_loop(self):

        while self.running:
            data = await self.reader.read(4)

            if len(data) != 4:
                continue
            read_len = unpack(">I", data)[0]

            data = await self.reader.read(read_len)

            response = make_msg()
            response.ParseFromString(data)
            resp = json.loads(response.payload_utf8)
            # print(response.namespace)
            # print(json.dumps(resp, indent=2))
            handler = self.dispatch.get(resp['type'])
            if handler:
                await handler(resp)

    async def start(self):
        self.reader, self.writer = \
            await asyncio.open_connection(self.host, self.port, ssl=ssl.SSLContext())

        # Connect and get-status
        await self.connect()
        await self.get_status()
        await self.ping()

        self.read_task = asyncio.create_task(self.recv_loop())

        while self.running:
            await asyncio.sleep(1)
            await self.stop()

    async def stop(self):
        if self.dmr_session:
            await self.stop_drm()
        await self.disconnect()

        self.running = False
        self.writer.close()
        await self.writer.wait_closed()

    async def send_message(self, ns, data, dst=None):

        msg = make_msg(ns)
        msg.payload_utf8 = json.dumps(data, ensure_ascii=False).encode('utf-8')
        if dst:
            msg.destination_id = dst

        sz = pack(">I", msg.ByteSize())
        self.writer.write(sz + msg.SerializeToString())
        await self.writer.drain()

    async def send_messager(self, ns, data, dst=None):
        self.id_counts[dst] += 1
        data['requestId'] = self.id_counts[dst]
        await self.send_message(ns, data, dst)

    async def dmr_message(self, data, ns='media'):
        assert self.dmr_session
        await self.send_messager(ns, data, dst=self.dmr_session)

    async def ping(self):
        await self.send_message('heartbeat', {"type": "PING"})

    async def pong(self, _):
        await self.send_message('heartbeat', {"type": "PONG"})

    async def get_status(self, dst=None):
        await self.send_messager('receiver', {"type": "GET_STATUS"}, dst)

    async def disconnect(self, dst=None):
        await self.send_message('con', {"type": "CLOSE", "origin": {}}, dst)

    async def connect(self, dst=None):
        await self.send_message('con', {"type": "CONNECT", "origin": {},
                                        "userAgent": "dmcast",
                                        "senderInfo": {}}, dst)

    async def launch_dmr(self):
        if self.dmr_session is None:
            await self.send_messager('receiver', {"type": "LAUNCH", "appId": DEFAULT_APP})
            await self.get_status()

    async def stop_drm(self):
        await self.dmr_message({"type": "STOP"}, 'receiver')
        self.dmr_session = None

    # Media

    async def media_load(self, media):
        assert self.dmr_session

        PLAY = {
            "type": "LOAD",
            "sessionId": self.dmr_session,
            "autoplay": True,
            "customData": {},
            "media": media}
        await self.dmr_message('media', PLAY)

    async def media_status(self):
        await self.dmr_message({"type": "GET_STATUS"})

    async def media_play(self, msid):
        await self.dmr_message({"type": "PLAY", "mediaSessionId": msid})

    async def media_pause(self, msid):
        await self.dmr_message({"type": "PAUSE", 'mediaSessionId': msid})

    async def media_stop(self, msid):
        await self.dmr_message({"type": "STOP", 'mediaSessionId': msid})

    # Device

    async def device_volume(self, volume):
        v = min(max(0, volume), 1)
        await self.send_messager('receiver', {"type": "SET_VOLUME", "volume": {"level": v}})

    async def device_mute(self, muted):
        await self.send_messager('receiver', {"type": "SET_VOLUME", "volume": {"muted": muted}})



# example media
# media = {
    # "contentId": "https://your-domain.com/hello.mp3",
    # "streamType": "BUFFERED",
    # "contentType": "audio/mp3",
    # "metadata": {}
# }
